// fpclassify.hpp

#ifndef BOOST_MATH_FPCLASSIFY_HPP
#define BOOST_MATH_FPCLASSIFY_HPP

// Copyright (c) 2006 Johan Rade

// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt
// or copy at http://www.boost.org/LICENSE_1_0.txt)

#include <cmath>
#include <cstring>
#include <boost/cstdint.hpp>
#include <boost/detail/endian.hpp>

#ifdef BOOST_PDP_ENDIAN
#	error PDP-endian platforms are not supported
#endif

namespace boost {
namespace math {

//------------------------------------------------------------------------------

namespace detail {

//------------------------------------------------------------------------------

// The following masks describe the meaning of the bits
// in the 4 most significant bytes of a floating point number

// long double

template<int is_extended> struct ld_mask
{
	static const boost::uint32_t sign;
	static const boost::uint32_t exponent;
	static const boost::uint32_t fraction;
};

// long double, sizeof(long double) == 8

template<> struct ld_mask<0>
{
	static const boost::uint32_t sign     = 0x80000000;
	static const boost::uint32_t exponent = 0x7ff00000;
	static const boost::uint32_t fraction = 0x000fffff;
};

// long double, sizeof(long double) > 8
//
// Most processors use the IEEE extended double format with 15 exponent bits,
// but Intel x86, Intel x64 and PowerPC have their own formats

template<> struct ld_mask<1>
{
	static const boost::uint32_t sign         = 0x80000000;

#if defined(__386) || defined(__i386__) \
	|| defined(_X86_) || defined(_M_IX86) \
	|| defined(__amd64) || defined(__amd64__)  || defined(_M_AMD64) \
	|| defined(__x86_64) || defined(__x86_64__) || defined(_M_X64)

	static const boost::uint32_t exponent    = 0x7fff0000;  
//	static const boost::uint32_t significand = 0x00008000;  
	static const boost::uint32_t fraction    = 0x00007fff;

#elif defined(__powerpc) || defined(__powerpc__) || defined(__POWERPC__) \
	|| defined(__ppc) || defined(__ppc__) || defined(__PPC__)

	static const boost::uint32_t exponent    = 0x7ff00000;  
	static const boost::uint32_t fraction    = 0x000fffff;

#else

	static const boost::uint32_t exponent    = 0x7fff0000;  
	static const boost::uint32_t fraction    = 0x0000ffff;

#endif
};

//..............................................................................

template<class T> struct mask
{
    static const boost::uint32_t sign;
	static const boost::uint32_t exponent;
	static const boost::uint32_t fraction;
};

template<> struct mask<float>
{
	static const boost::uint32_t sign     = 0x80000000;
	static const boost::uint32_t exponent = 0x7f800000;
	static const boost::uint32_t fraction = 0x007fffff;
};

template<> struct mask<double>
{
	static const boost::uint32_t sign     = 0x80000000;
	static const boost::uint32_t exponent = 0x7ff00000;
	static const boost::uint32_t fraction = 0x000fffff;
};

template<> struct mask<long double>
{
	static const int is_extended = sizeof(long double) > 8;

	static const boost::uint32_t sign     = ld_mask<is_extended>::sign;
	static const boost::uint32_t exponent = ld_mask<is_extended>::exponent;
	static const boost::uint32_t fraction = ld_mask<is_extended>::fraction;
};

//------------------------------------------------------------------------------

// The function copy copies the four most significant bytes
// of a floating point number

//..............................................................................

#if defined BOOST_BIG_ENDIAN

template<class T> void copy(T x, boost::uint32_t& a )
{
	memcpy(&a, &x, 4);
}

template<class T> void copy(boost::uint32_t a, T& x)
{
	memcpy(&x, &a, 4);
}

//..............................................................................

#elif defined BOOST_LITTLE_ENDIAN

template<class T> void copy(T x, boost::uint32_t& a )
{
	memcpy(
		&a,
		reinterpret_cast<const unsigned char*>(&x) + sizeof(T) - 4,
		4
	);
}

template<class T> void copy(boost::uint32_t a, T& x)
{
	memcpy(
		reinterpret_cast<unsigned char*>(&x) + sizeof(T) - 4,
		&a,
		4
	);
}

#if defined(__386) || defined(__i386__) \
	|| defined(_X86_) || defined(_M_IX86) \
	|| defined(__amd64) || defined(__amd64__)  || defined(_M_AMD64) \
	|| defined(__x86_64) || defined(__x86_64__) || defined(_M_X64)

// On Intel x86 and Intel x64, when sizeof(long double) > 8,
// then there are only 10 significant bytes.
// The rest is padding. Then we replace sizeof(T) - 4 by 10 - 4 = 6.

inline void copy(long double x, boost::uint32_t& a )
{
	memcpy(
		&a,
		reinterpret_cast<const unsigned char*>(&x)
			+ (sizeof(long double) == 8 ? 4 : 6),
		4
	);
}

inline void copy(boost::uint32_t a, long double& x)
{
	memcpy(
		reinterpret_cast<unsigned char*>(&x)
			+ (sizeof(long double) == 8 ? 4 : 6),
		&a,
		4
	);
}

#endif

//..............................................................................

#else
#	error We should not get here
#endif

//------------------------------------------------------------------------------

}   // namespace detail

//------------------------------------------------------------------------------

template<class T> bool (isfinite)(T x)
{
    boost::uint32_t a;
    detail::copy(x,a);
    a &= detail::mask<T>::exponent;
    return a != detail::mask<T>::exponent;
}

//..............................................................................

template<class T> bool (isnormal)(T x)
{
	boost::uint32_t a;
	detail::copy(x,a);
	a &= detail::mask<T>::exponent;
	return (a != 0) && (a != detail::mask<T>::exponent);
}

//..............................................................................

template<class T> bool (isinf)(T x)
{
	boost::uint32_t a;
	detail::copy(x,a);
	if((a & detail::mask<T>::exponent) != detail::mask<T>::exponent)
		return false;
	a &= detail::mask<T>::fraction;
	detail::copy(a,x);
	return x == 0;
}

inline bool (isinf)(float x)
{
    boost::uint32_t a;
    detail::copy(x,a);
    a &= ~detail::mask<float>::sign;
    return a == detail::mask<float>::exponent;
}

//..............................................................................

template<class T> bool (isnan)(T x)
{
	boost::uint32_t a;
	detail::copy(x,a);
	if((a & detail::mask<T>::exponent) != detail::mask<T>::exponent)
		return false;
	a &= detail::mask<T>::fraction;
	detail::copy(a,x);
	return x != 0;
}

inline bool (isnan)(float x)
{
	boost::uint32_t a;
	detail::copy(x,a);
	return (a & detail::mask<float>::exponent) == detail::mask<float>::exponent
		   && (a & detail::mask<float>::fraction) != 0;
}

//------------------------------------------------------------------------------

#ifndef FP_INFINITE
#   define FP_INFINITE 0
#   define FP_NAN 1
#   define FP_NORMAL 2
#   define FP_SUBNORMAL 3
#   define FP_ZERO 4
#endif

template<class T> int (fpclassify)(T x)
{
    boost::uint32_t a;
    detail::copy(x,a);
    a &= detail::mask<T>::exponent;

    if(a == 0) {
        if(x == 0)
			return FP_ZERO;
		else
            return FP_SUBNORMAL;
    }

    else if(a == detail::mask<T>::exponent) {
        if((isinf)(x))								// This can be optimized
			return FP_INFINITE;
        else
			return FP_NAN;
    }

    else
        return FP_NORMAL;
}

//------------------------------------------------------------------------------

template<class T> bool (signbit)(T x)
{
    boost::uint32_t a;
    detail::copy(x,a);
    a &= detail::mask<T>::sign;
    return a != 0;
}

template<class T> T (copysign)(T x, T y)      // magnitude of x and sign of y
{
    boost::uint32_t a;
    detail::copy(x,a);
    a &= ~detail::mask<T>::sign;

    boost::uint32_t b;
    detail::copy(y,b);
    b &= detail::mask<T>::sign;

    detail::copy(a|b,x);
    return x;
}

template<class T> T (changesign)(T x)
{
    boost::uint32_t a;
    detail::copy(x,a);
    a ^= detail::mask<T>::sign;
    detail::copy(a,x);
    return x;
}

//------------------------------------------------------------------------------

}   // namespace math
}   // namespace boost

#endif
