// test.cpp

// Copyright (c) 2006 Johan Rade

// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt
// or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifdef _MSC_VER
#   pragma warning(disable: 4127 4244 4511 4512 4701 4800)
#endif

//-------------------------------------

#include "../fpclassify.hpp"

#undef NDEBUG

#include <cassert>
#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <limits>

//------------------------------------------------------------------------------

// The anonymous namespace ensures that there are no ambiguities between
// boost::math::fpclassify etc and std::fpclassify etc

namespace {

//------------------------------------------------------------------------------

using boost::math::fpclassify;
using boost::math::isfinite;
using boost::math::isnormal;
using boost::math::isinf;
using boost::math::isnan;

using boost::math::signbit;
using boost::math::copysign;
using boost::math::changesign;

//------------------------------------------------------------------------------

template<class T> void test_basic();
template<class T> void test_copysign();
template<class T> void test_changesign();

template<class T> void print(const char* s, T x, bool ok = true);


//------------------------------------------------------------------------------

}	// namespace

//------------------------------------------------------------------------------

int main()
{
	std::cout << '\n';

#if defined(__386) || defined(__i386__) \
	|| defined(_X86_) || defined(_M_IX86) \

	std::cout << "x86\n";
	
#elif defined(__amd64) || defined(__amd64__)  || defined(_M_AMD64) \
	|| defined(__x86_64) || defined(__x86_64__) || defined(_M_X64)

	std::cout << "x64\n";

#elif defined(__powerpc) || defined(__powerpc__) || defined(__POWERPC__) \
	|| defined(__ppc) || defined(__ppc__) || defined(__PPC__)

	std::cout << "Power PC\n";

#else

	std::cout << "Other\n";

#endif

	//......................

#if defined BOOST_BIG_ENDIAN
	std::cout << "big endian\n";
#elif defined BOOST_LITTLE_ENDIAN
	std::cout << "little endian\n";
#else
#	error should not get here
#endif

	print(
		"long double inf = ",
		std::numeric_limits<long double>::infinity(),
		std::numeric_limits<long double>::has_infinity);

	//.....................

	std::cout << "Testing float" << std::endl;
    test_basic<float>();
	test_copysign<float>();
	test_changesign<float>();

	std::cout << "Testing double" << std::endl;
	test_basic<double>();
    test_copysign<double>();
	test_changesign<double>();

	std::cout << "Testing long double" << std::endl;
	test_basic<long double>();
    test_copysign<long double>();
    test_changesign<long double>();

	std::cout << std::endl;

    return 0;
}

//------------------------------------------------------------------------------

namespace {

//------------------------------------------------------------------------------

template<class T>
void test_basic()
{
    assert(std::numeric_limits<T>::has_infinity);
	assert(std::numeric_limits<T>::has_quiet_NaN);
	assert(std::numeric_limits<T>::has_signaling_NaN);
    T x;

    x = static_cast<T>(0);
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_ZERO);

    x = (changesign)(static_cast<T>(0));
    assert((signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_ZERO);

    x = std::numeric_limits<T>::denorm_min();
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_SUBNORMAL);

    x = -std::numeric_limits<T>::denorm_min();
    assert((signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_SUBNORMAL);
/*
    // fail on some platforms due to incorrect definition of 
    // std::numeric_limits<long double>::min()

    x = std::numeric_limits<T>::min() / 2;
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_SUBNORMAL);

    x = -std::numeric_limits<T>::min() / 2;
    assert((signbit)(x));
    assert((isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_SUBNORMAL);
*/
    x = (std::numeric_limits<T>::min)();
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = -(std::numeric_limits<T>::min)();
    assert((signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = static_cast<T>(1);
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = static_cast<T>(-1);
    assert((signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = (std::numeric_limits<T>::max)();
    assert(!(signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = -(std::numeric_limits<T>::max)();
    assert((signbit)(x));
    assert((isfinite)(x));
    assert((isnormal)(x));
    assert(!(isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_NORMAL);

    x = std::numeric_limits<T>::infinity();
    assert(!(signbit)(x));
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert((isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_INFINITE);

    x = -std::numeric_limits<T>::infinity();
    assert((signbit)(x));
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert((isinf)(x));
    assert(!(isnan)(x));
    assert((fpclassify)(x) == FP_INFINITE);

    x = std::numeric_limits<T>::quiet_NaN();
    //assert(!(signbit)(x));				// funny <limits> on some platforms
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert((isnan)(x));
    assert((fpclassify)(x) == FP_NAN);

    x = (changesign)(std::numeric_limits<T>::quiet_NaN());
    //assert((signbit)(x));				// funny <limits> on some platforms
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert((isnan)(x));
    assert((fpclassify)(x) == FP_NAN);

    x = std::numeric_limits<T>::signaling_NaN();
    //assert(!(signbit)(x));				// funny <limits> on some platforms
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert(!(isinf)(x));
    assert((isnan)(x));
    assert((fpclassify)(x) == FP_NAN);

    x = (changesign)(std::numeric_limits<T>::signaling_NaN());
    //assert((signbit)(x));				// funny <limits> on some platforms
    assert(!(isfinite)(x));
    assert(!(isnormal)(x));
    assert((isnan)(x));
    assert((fpclassify)(x) == FP_NAN);
}

//------------------------------------------------------------------------------

#define countof(A) (sizeof(A)/sizeof((A)[0]))

template<class T> void test_copysign_impl(T x, T y);

template<class T> void test_copysign()
{
    T a[] = {
        static_cast<T>(0),
        std::numeric_limits<T>::denorm_min(),
        (std::numeric_limits<T>::min)() / 2,
        (std::numeric_limits<T>::min)(),
        static_cast<T>(1),
        (std::numeric_limits<T>::max)(),
        std::numeric_limits<T>::infinity(),
    };

    for(int i = 0; i < countof(a); ++i) {
        for(int j = 0; j < countof(a); ++j)
            test_copysign_impl(a[i], a[j]);
    }
}

template<class T> void test_copysign_impl(T x, T y)
{
    assert((copysign)(x, y) == x);
    assert((copysign)(x, (changesign)(y)) == (changesign)(x));
    assert((copysign)((changesign)(x), y) == x);
    assert((copysign)((changesign)(x), (changesign)(y)) == (changesign)(x));
}

//------------------------------------------------------------------------------

template<class T> void test_changesign()
{
    T a[] = {
        std::numeric_limits<T>::denorm_min(),
        (std::numeric_limits<T>::min)(),
        static_cast<T>(1),
        std::numeric_limits<T>::infinity()
    };

    for(int i = 0; i < countof(a); ++i) {
        assert((changesign)(a[i]) == -a[i]);
        assert((signbit)((changesign)(a[i])));
        assert((changesign)(-a[i]) == a[i]);
        assert(!(signbit)((changesign)(-a[i])));
    }

    T b = static_cast<T>(0);
    assert((changesign)(b) == b);
    assert((signbit)((changesign)(b)));
    assert((changesign)((changesign)(b)) == b);
    assert(!(signbit)((changesign)((changesign)(b))));

    T c = std::numeric_limits<T>::quiet_NaN();
    assert((isnan)((changesign)(c)));
    //assert((signbit)((changesign)(c)));				// funny <limits> on some platforms
    assert((isnan)((changesign)((changesign)(c))));
    //assert(!(signbit)((changesign)((changesign)(c))));	// funny <limits> on some platforms
}

//------------------------------------------------------------------------------

template<class T>
void print(const char* s, T x, bool ok)
{

	std::cout << std::left << std::setw(10) << s << std::right;

	std::cout << std::hex << std::setfill('0');

	if(ok) {
#ifdef BOOST_BIG_ENDIAN
		for(int i = 0; i < (int)sizeof(T); ++i)
#else
		for(int i = sizeof(T) - 1; i >= 0; --i)
#endif
			std::cout << std::setw(2) << (unsigned int)*((unsigned char*)&x + i) << ' ';
	}
	else {
		for(int i = 0; i < (int)sizeof(T); ++i)
			std::cout << "-- ";
	}

	std::cout << '\n';
	std::cout << std::dec << std::setfill(' ');
}

//------------------------------------------------------------------------------

}	// namespace

